/**********************************************************
 **
 ** operator.c : manipulators for operator_t
 **
 ** Copyright (c) 2004-2005, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include "operator.h"

result_t operator_exec(operator_t op, variable_const_t *dst,
    variable_const_t a, variable_const_t b)
{
  result_t rc;
  
  rc = result_ok;
  switch (op) {
    case operator_neg:
      a = -a;
      break;
    case operator_add:
      a += b;
      break;
    case operator_sub:
      a -= b;
      break;
    case operator_mul:
      a *= b;
      break;
    case operator_div:
      if (!b) {
        rc = result_divide_by_zero;
      } else {
        a /= b;
      }
      break;
    case operator_mod:
      if (!b) {
        rc = result_divide_by_zero;
      } else {
        a = a % b;
      }
      break;
    case operator_lt:
      a = a < b;
      break;
    case operator_le:
      a = a <= b;
      break;
    case operator_eq:
      a = a == b;
      break;
    case operator_ne:
      a = a != b;
      break;
    case operator_ge:
      a = a >= b;
      break;
    case operator_gt:
      a = a > b;
      break;
    case operator_andl:
      a = a && b;
      break;
    case operator_orl:
      a = a || b;
      break;
    case operator_notl:
      a = !a;
      break;
    case operator_andb:
      a = a & b;
      break;
    case operator_orb:
      a = a | b;
      break;
    case operator_xorb:
      a = a ^ b;
      break;
    case operator_cmpb:
      a = ~a;
      break;
    case operator_assign:
      break; /* return A */
    case operator_null:
      rc = result_syntax;
      break;
    case operator_shift_left:
      a <<= b;
      break;
    case operator_shift_right:
      a >>= b;
      break;
    case operator_logical:
      a = !!a;
      break;
    case operator_incr:
      a++;
      break;
    case operator_decr:
      a--;
      break;
    case operator_dot:
    case operator_subscript:
      break;
  }
  if (result_ok == rc) {
    *dst = a;
  }
  return rc;
}

/*
 * NAME
 *   operator_is_unary
 *
 * DESCRIPTION
 *   determine if an operator is a unary op
 *
 * PARAMETERS
 *   op : operator
 *
 * RETURN
 *   0 : operator is not unary
 *   1 : operator is unary
 *
 * NOTES
 */
boolean_t operator_is_unary(operator_t op)
{
  boolean_t rc;

  rc = boolean_false;
  switch (op) {
    case operator_null:
    case operator_add:
    case operator_sub:
    case operator_mul:
    case operator_div:
    case operator_mod:
    case operator_lt:
    case operator_le:
    case operator_eq:
    case operator_ne:
    case operator_ge:
    case operator_gt:
    case operator_andl:
    case operator_orl:
    case operator_andb:
    case operator_orb:
    case operator_xorb:
    case operator_shift_left:
    case operator_shift_right:
    case operator_assign:
    case operator_dot:
    case operator_subscript:
      break;
    case operator_logical:
    case operator_notl:
    case operator_cmpb:
    case operator_neg:
    case operator_incr:
    case operator_decr:
      rc = boolean_true;
      break;
  }
  return rc;
}

boolean_t operator_is_assign(operator_t op)
{
  boolean_t rc;

  rc = boolean_false;
  switch (op) {
    case operator_null:
    case operator_add:
    case operator_sub:
    case operator_mul:
    case operator_div:
    case operator_mod:
    case operator_lt:
    case operator_le:
    case operator_eq:
    case operator_ne:
    case operator_ge:
    case operator_gt:
    case operator_andl:
    case operator_orl:
    case operator_andb:
    case operator_orb:
    case operator_xorb:
    case operator_shift_left:
    case operator_shift_right:
    case operator_logical:
    case operator_notl:
    case operator_cmpb:
    case operator_neg:
    case operator_incr:
    case operator_decr:
    case operator_dot:
    case operator_subscript:
      break;
    case operator_assign:
      rc = boolean_true;
      break;
  }
  return rc;
}

boolean_t operator_is_binary(operator_t op)
{
  return !(operator_is_unary(op) || operator_is_assign(op));
}

boolean_t operator_is_logical(operator_t op)
{
  boolean_t rc;

  rc = boolean_false;
  switch (op) {
    case operator_null:
    case operator_add:
    case operator_sub:
    case operator_mul:
    case operator_div:
    case operator_mod:
    case operator_lt:
    case operator_le:
    case operator_eq:
    case operator_ne:
    case operator_ge:
    case operator_gt:
    case operator_andb:
    case operator_orb:
    case operator_xorb:
    case operator_shift_left:
    case operator_shift_right:
    case operator_cmpb:
    case operator_neg:
    case operator_incr:
    case operator_decr:
    case operator_assign:
    case operator_dot:
    case operator_subscript:
      break;
    case operator_andl:
    case operator_orl:
    case operator_logical:
    case operator_notl:
      rc = boolean_true;
      break;
  }
  return rc;
}

boolean_t operator_is_relation(operator_t op)
{
  boolean_t rc;

  rc = boolean_false;
  switch (op) {
    case operator_null:
    case operator_add:
    case operator_sub:
    case operator_mul:
    case operator_div:
    case operator_mod:
    case operator_andl:
    case operator_orl:
    case operator_andb:
    case operator_orb:
    case operator_xorb:
    case operator_shift_left:
    case operator_shift_right:
    case operator_logical:
    case operator_notl:
    case operator_cmpb:
    case operator_neg:
    case operator_incr:
    case operator_decr:
    case operator_assign:
    case operator_dot:
    case operator_subscript:
      break;
    case operator_lt:
    case operator_le:
    case operator_eq:
    case operator_ne:
    case operator_ge:
    case operator_gt:
      rc = boolean_true;
      break;
  }
  return rc;
}

boolean_t operator_is_commutative(operator_t op)
{
  boolean_t rc;

  rc = boolean_false;
  switch (op) {
    case operator_null:
    case operator_sub:
    case operator_div:
    case operator_mod:
    case operator_shift_left:
    case operator_shift_right:
    case operator_logical:
    case operator_notl:
    case operator_cmpb:
    case operator_neg:
    case operator_incr:
    case operator_decr:
    case operator_assign:
    case operator_lt:
    case operator_le:
    case operator_ge:
    case operator_gt:
    case operator_dot:
    case operator_subscript:
      break;
    case operator_add:
    case operator_mul:
    case operator_andl:
    case operator_orl:
    case operator_andb:
    case operator_orb:
    case operator_xorb:
    case operator_eq:
    case operator_ne:
      rc = boolean_true;
      break;
  }
  return rc;
}

